/* Instructables
 *  Project Eagle I
 *  Wireless_Sentry_Defence_
 *  
 *  Authors: Shafae Ali, Oscar Jones
 *  
 *  UNO File
 *  The project utilises an Arduino UNO (which controls the turret) and an Arduino Nano (which controls the joystick).
 *  Both Arduinos are wirelessly communicating with each other on a 2.4GHz wireless network, resulting in two seperate .ino files.
 *  One is for inputting commands (the joystick/Nano), the other is for outputting the data sent to it (the turret/UNO).
 *  The following code is for the turret and is used on the Arduino UNO.
 *  One key note for the script is that care has been taken to avoid any lines of code that cause interuptions or delays in other commands taking place.
 *  The script has been written with the intention to focus on rapid response, given than it is operating wirelessly.
 *  The implementation reflects this as in actual use, there is negligible delay between turret outcome and joystick interaction.
 */

//the SPI library is used for communicating with the NRF24L01 module. 
//It uses the SCK, MISO and MOSI pins for communicating quickly over short distances, e.g. wireless module, RTC module
#include <SPI.h>
//nRF24L01 and RF24 are libraries for using the wireless module
#include <nRF24L01.h>
#include <RF24.h>
//the Servo libary makes including multiple servos compact and provides the required signal through simple commands
#include <Servo.h>

//usage of the Servo libaray, defining firstly the 3 servo units angle_servo(the tilt on the turret),continous_servo(turret swivel)
//and the servo_feed (mounted on the magazine for feeding balls into the barrel)
//The ESCs makes use of the Servo library as they can be controlled through 1 data pin, it only requires a PWM signal to operate similarly to a servo
//The main difference when using ESCs is that the ESC must receive a low PWM signal to arm itself,without this inital arming process upon turn on, the ESC will not respond.
//The ESC used is configured in such a way that it must start in neutral position, therefore the Arduino must deliver a 1500 microsecond signal.
// This is because the Arduino writes between 1000 (considered off for the ESC) to 2000(max)
//As a result, esc1 and esc2 have been defined as servos here as they will be driving the brushless motors attached to the rotating wheels.
Servo angle_servo, continous_servo, servo_feed, esc1, esc2;

//this is to define the wireless module pins seperately to the rest of the code. This makes redfining them simple as it is core to our data transmission
//the CE_PIN is used for data transmission and is the primary pin for the wireless module
#define CE_PIN 8
//the CSN pin is an active low pin for the SPI bus
#define CSN_PIN 10

//an address for communicating over is established as a read-only variable using const
//a random selection of letters was chosen as a unqiue address
const byte address[5] = {'R', 'x', 'A', 'A', 'A'};

//use of RF24 in which the two non-standard pins are defined for radio communication
//SCK, MOSI and MISO are pre-defined as pin 13,12 and 11, whereas CE and CSN are chosen
RF24 radio(CE_PIN, CSN_PIN);

//a single array that holds 4 values which will be read from the incoming Nano signal
//This is more efficient and robust than trying to send 4 seperate values, instead a single message is recieved
int message[4];

//Groupings of the component pin and associated data being used with it have been made

//defining the tilt servo to a digital pin
//producing an integer that will hold the potentiometer value recieved from the Nano
int angle_servo_pin = 9;
int angle_val;

//defining the swivel servo to a digital pin
//producing an integer that will hold the potentiometer value recieved from the Nano
//the 360 continous rotation servo does not linearly scale, therefore the map function cannot be used to convert the rotation value directly to a direction
//as a result a sepearte rotation direction variable has to be made to tackle this later in the code
int continous_servo_pin = 4;
int rotation_val;
int rotation_direction;

//defining the feeding servo to a digital pin
//producing an integer that holds an on or off for firing state
int servo_feed_pin = 7;
int firing_state;

//defining each ESC to a digital pin
//since both ESCs should be rotating at the same speed, a single escSpeed integer is used that will write to both ESC, and therefore motor speeds are matched
int esc1_pin = 6;
int esc2_pin = 5;
int escSpeed;

//defining the relay digital control pin to an UNO digital pin
int relay_pin = 2;

/////////////////////////////////////////////////////////

void setup() {

  //this was initally used in setup during troubleshooting and iteratively developing code whilst interfacing to the hardware
  //it has been commented out for now, but has been kept for quick development requiring for reading values, etc.
  //Serial.begin(9600);

  //this starts radio transmission and utilises the RF24 module
  radio.begin();
  //the lowest data transfer rate provides greater connection range
  //since a single array is being received, this rate is suitable as the more important factor is the wireless range
  radio.setDataRate( RF24_250KBPS );
  //the nRF24L01 has 6 pipes that it can transmit or receive data to at a time
  //this command opens the first data pipe for reading at the address we defined earlier
  radio.openReadingPipe(0, address);
  //Since the UNO is only receiving data, it must start listening for any values coming through the aforementioned ReadingPipe
  radio.startListening();

  //attach is a Servo library command that allows the earlier defined servo to be given a corresponding pin
  //The code has been structured to group the three servos together
  angle_servo.attach(angle_servo_pin);
  continous_servo.attach(continous_servo_pin);
  servo_feed.attach(servo_feed_pin);
  
  //the following group is for the ESCs
  esc1.attach(esc1_pin);
  esc2.attach(esc2_pin);

  //the relay is a simple component where either a high or low is sent to it to activate it, therefore it is set as a digital output in setup
  pinMode(relay_pin, OUTPUT);
  
}

/////////////////////////////////////////////////////////

void loop() {

  //to ensure that the UNO only responds if there is a radio signal, it will only commence the loop function when a connection has been establised to the transmitter (in this case the Nano)
  //from here on in it is assumed that communication is established and the radio can be used
  if ( radio.available() ) {

    ////////////////RADIO MESSAGE RECEIVING//////////////////
    //'sizeof' is a command for writing the message to the address, the second argument confirms the number of bytes occupied in the array and how many should be received
    //the array defined at the start is being read, so if a 4 part array is being sent from the Nano, it will be filled into the one defined here
    //using '&' in this instance will pass the address of the variable where the data is stored
    radio.read( &message, sizeof(message) );

    //now that the message is received, each part of the array will be defined to a variable that can be used to write to the output components
    //identical variable names have been used across the Nano and Uno so that it is clear to understand which values correspond to each other
    //each variable name is self-identifying to the turret
    angle_val = message[0];
    rotation_val = message[1];
    escSpeed = message[2];
    firing_state = message[3];
    
    ////////////////TURRET TILT//////////////////
    //the remapping of this value from potentiometer input to suitable values for a servo was done on the Nano
    //the command here simply writes the angle value to the servo 
    //this had to be fine tuned through testing to find the max tilt and min tilt
    //the output results from the input that corresponds to usable angles
    angle_servo.write(angle_val);
    
    ////////////////TURRET SWIVEL//////////////////
    //as mentioned previously the continous servo does not scale linearly and it has a deadzone in which it will not rotate, but go to a fixed location, therefore it must be controlled using a set of if statements
    //it has a zone that angles can be written to (49 to 97) and the servo will go to that location.
    //Anything above 97 or below 49 will result in continous rotation in that direction.
    //the reason it is included on the UNO is to group processing of the code being sent from the Nano together.
    //The code takes the potentiometer value and initiates an if statement that writes a speed to the servo depending on how far the joystick is tilted
    //LEFT ON THE JOYSTICK CAUSES COUNTERCLOCKWISE ROTATION
    if (rotation_val < 70) {
      rotation_direction = 100;
      //the attach line is crucial and is used due to the detach line used later
      //the reason for this is explained in the detach line
      continous_servo.attach(continous_servo_pin);
      continous_servo.write(rotation_direction);
    }
    //this if statement causes the turret to rotate faster if the joystick is pushed further to the left.
    if (rotation_val < 55) {
      rotation_direction = 110;
      continous_servo.attach(continous_servo_pin);
      continous_servo.write(rotation_direction);
    }
    //RIGHT ON THE JOYSTICK CAUSES CLOCKWISE ROTATION
    if (rotation_val > 120) {
      rotation_direction = 47;
      continous_servo.attach(continous_servo_pin);
      continous_servo.write(rotation_direction);
    }

    //when the joystick returns to neutral position, the potentiometer is at a fixed value.
    //This value coresponds to a value for the continous servo where it can spring back rather than carry on rotating
    //as a result the detach command is used to disable the pin that the servo is connected to such that, only when the joystick is beyond a certain point going left or right, will the servo rotate
    if (rotation_val <= 120 && rotation_val >= 70) {
      continous_servo.detach();
    }
    
    ////////////////TURRET FIRING & SAFETY//////////////////
    //the following if statement provides a safety latch where the && command ensures only when the wheels are spinning beyond a minimum value (this is if we want to increase that value on the Nano side of the code)and the button is pressed, does the servo feed any balls to the barrel
    //N.B. The state of the firing pin is LOW when pressed and HIGH when not pressed
    if (firing_state == LOW && escSpeed>=1550) {
      servo_feed.write(125);
    }
    //when the button is not pressed the servo is blocking the feed of the balls
    if (firing_state == HIGH) {
      servo_feed.write(160);
    }
    
    ////////////////TURRET WHEEL SPEED//////////////////
    //ESC understands commands in microSeconds and this writes whichever values are set by the Nano
    //The ESCs are constantly running on idle speed. When the button on the joystick is pressed, then a new value is continously written to the ESC
    //There is minimal processing of the ESC code on the UNO as the button is on the joystick, making it easier to process it there
    //this line continously writes the data to the ESC, which is required otherwise the ESC loses it's netural throttle position
    esc1.writeMicroseconds(escSpeed);
    esc2.writeMicroseconds(escSpeed);
    
    ////////////////BLOWER FAN ////////////////////////
    //if the ESC is neutral and off, the relay is HIGH corresponding to fans being off
    if (escSpeed == 1500) {
      digitalWrite(relay_pin, HIGH);
    }
    //if the ESC is spinning, the relay switches to LOW corresponding to the fans being on
    if (escSpeed >= 1550) {
      digitalWrite(relay_pin, LOW);
    }

  }
}
